﻿# -*- coding: utf-8 -*-
#
# methods to calculate Q1 and Q2 of Yanai et al. 1973
#
#	Eriko Nishimoto
#
=begin
= Methods to calculate Q1 and Q2
== module GPhys_Function
    The following functions were made by eriko.

    You need GPhys libraries.

=== Module Functions
--- derive_q1( t, gph, sh, u, v, omega )

    calculating Q1 (apparent heat source)

    ARGUMENTS
    * t      : temperature [K] (GPhys)
    * gph    : geopotential height [m2/s2] (GPhys)
    * sh     : specific humidity [kg/kg] (GPhys)
    * u      : zonal wind velocity [m/s] (GPhys)
    * v      : meridional wind velocity [m/s] (GPhys)
    * omega  : pressure vertical velocity [Pa/s] (GPhys)

    RETURN VALUE
    * q1     : apparent heat source (GPhys)
    * s      : dry static energy (GPhys)
    * ds_dt  : ds/dt (GPhys)
    * div_s  : horizontal divergence of s (GPhys)
    * w_ds_dp: vertical advection of s (GPhys)

--- derive_q2( sh, u, v, omega )

    calculating Q2 (apparent moisture sink)

    ARGUMENTS
    * sh     : specific humidity [kg/kg] (GPhys)
    * u      : zonal wind velocity [m/s] (GPhys)
    * v      : meridional wind velocity [m/s] (GPhys)
    * omega  : pressure vertical velocity [Pa/s] (GPhys)

    RETURN VALUE
    * q2     : apparent heat source (GPhys)
    * dsh_dt : time derivative of specific humidity (GPhys)
    * div_sh : horizontal divergence of sh (GPhys)
    * w_dh_dp: vertical advection of sh (GPhys)


=end

require "numru/ggraph"
include NumRu

require "numru/gphys/derivative"
require "numru/ganalysis/planet"

module GPhys_Function

  DCL.gllset "lmiss",true
  @rmiss = 999.0
  DCL.glrset "rmiss",@rmiss

  #< Themodynamic constants for the Earth's atmosphere >

  Cp  = UNumeric.new(1004.6,"J/kg/K")    # dry isobaric specific heat
  @@g = UNumeric.new(9.8,"m.s-2")        # gravity acceleration
  Lc  = UNumeric.new(2.5E+6,"J/kg")      # latent heat

  module_function

  #< module parameters >

  # GPhysの軸の単位を変換する
  #
  # ARGUMENT
  # * data [GPhys]
  # * dim [integer or character] : the dimention of which the unit is converted.
  # * units [character]
  # EXAMPLE
  #   GPhys_Function::convert_axis_units( data, "pres", "hPa" )
  # RETURN VALUE
  # * data [GPhys]
  def convert_axis_units( data, dim, units )
    z1 = data.axis(dim).pos.convert_units(Units[units])
    data.axis(dim).set_pos(z1)

    return data
  end

  # GPhysの圧力座標軸の単位を(millibarから)hPaに変換する
  #
  # ARGUMENT
  # * data [GPhys which contains a pressure coordinate]
  # * dim [Integer or character] : the dimention of the pressure coordinate
  # EXAMPLE
  #   GPhys_Function::convert_axis_units2hPa( data, "pres" )
  # RETURN VALUE
  # * data [GPhys]
  def convert_axis_units2hPa( data, dim=2 )
    unit_org = data.coord(dim).units
    if Units["hPa"] == unit_org
    elsif Units["Pa"] == unit_org
      data = convert_axis_units( data, dim=2, units="hPa" )
    elsif /^millibar/ =~ unit_org.to_s
      z1 = data.axis(dim).pos
      z1.units = Units["hPa"]
      data.axis(dim).set_pos(z1)
    else
      p "#{unit_org.to_s} is not supported in GPhys_Function::convert_axis_units2hPa"
      exit
    end

    return data
  end

  # GPhysの軸の値をソートする
  #
  # ARGUMENT
  # * data [GPhys]
  # * dim [Integer or character]
  # EXAMPLE
  #   GPhys_Function::sort_ax( data, "lat" )
  # RETURN VALUE
  # * data [GPhys]
  def sort_ax( data, dim=1 )
    ax_val = data.coord(dim).val

    if (ax_val[1] - ax_val[0]) < 0
      sortindex = ax_val.sort_index

      new = data[*([true]*dim + sortindex + [false])]
      ax_va = new.coord(dim).replace_val( ax_val.sort )

      new.axis(dim).set_pos( ax_va )
      return new
    else
      return data
    end
  end

  # GPhysデータの単位を変換する
  #
  # ARGUMENT
  # * data [GPhys]
  # * vartype [character] : should be t, u, v, gph, omega, specific_humidity
  # EXAMPLE
  #   GPhys_Function::convert_var_units( data, "t" )
  # RETURN VALUE
  # * data [GPhys]
  def convert_var_units( data, vartype )
    unit_org = data.units

    unit_list = {"t"=>"K","u"=>"m/s","v"=>"m/s"\
      ,"gph"=>"m2/s2","omega"=>"hPa/s"\
      ,"specific_humidity"=>"kg/kg" }

    data = data.convert_units( Units[ unit_list[ vartype ] ] )

    return data
  end

  # ジオポテンシャルハイトからジオポテンシャルを計算する
  #
  # ARGUMENT
  # * gph [GPhys] Geopotential Height (m)
  # RETURN VALUE
  # * data [GPhys] Geopotential -- g*gph (m2/s2)
  def get_geopotential( gph )
    unit_org = gph.units
    if Units["m2/s2"] === unit_org
      return gph
    elsif Units["m"] === unit_org
      return @@g*gph
    else
      p "#{unit_org} is not supported in GPhys_Function::get_geopotential"
      exit
    end
  end

  # Q1, Q2を計算するにあたってのデータの前処理
  #
  # ARGUMENT
  # * data [GPhys]
  # * vartype [character] : should be t, u, v, gph, omega, specific_humidity
  # RETURN VALUE
  # * data [GPhys]
  def prep_deviation( data, vartype )
    tmp = data.copy
    tmp = convert_axis_units2hPa( tmp, dim=2 )
    tmp = sort_ax( tmp, dim=1 )
    if vartype == "gph"
      tmp = get_geopotential( tmp )
    else
      tmp = convert_var_units( tmp, vartype )
    end

    return tmp
  end

  def time2sec( data, dim=-1 )
    timeunitsorg = data.coord(dim).units
    timeval = data.coord(dim).val

    case timeunitsorg.to_s
    when /^seconds/, "s"
      timefactor=1
      z1val = (timeval-timeval[0]) * timefactor
    when /^minits/
      timefactor=60
      z1val = (timeval-timeval[0]) * timefactor
    when /^hours/
      timefactor=60*60
      z1val = (timeval-timeval[0]) * timefactor
    when /^days/
      timefactor=60*60*24
      z1val = (timeval-timeval[0]) * timefactor
    when /^month/
      calendar = data.coord(dim).get_att("calendar")
      calendar = "gregorian" if calendar==nil

      st = UNumeric::new(timeval[0], timeunitsorg).to_datetime( 0.1, calendar )
      z1val = NArray.int( timeval.length )
      tti=0
      timeval.each{|tt|
        unumeric_time = UNumeric::new(tt, timeunitsorg)
        tstr = unumeric_time.to_datetime( 0.1, calendar )
        z1val[tti] = ( tstr - st ).to_i * 60 * 60 * 24
        tti+=1
      }
    else
      p "#{timeunitsorg.to_s} is not supported in GPhys_Function::time2sec"
      exit
    end
    timealt = VArray.new( z1val, {"units"=>"s"}, "timealt" )

    return timealt
  end

  def total_differentiation( s, u, v, omega )

    timealt = time2sec( s )
    p timealt
    #p timealt.units
    ds_dt = GPhys::Derivative::cderiv( s, -1, bc=1, timealt )
    ds_dt.rename("d#{s.name}_dt")
    ds_dt.long_name = "Tendency of #{s.name}"

    ds_dx, ds_dy = GAnalysis::Planet::grad_s( s )
    div_s = u*ds_dx + v*ds_dy
    div_s.rename("div_#{s.name}")
    div_s.long_name = "Horizontal Advection of #{s.name}"

    ds_dp = GPhys::Derivative::cderiv( s, 2, bc=1 )
    w_ds_dp = omega*ds_dp
    w_ds_dp.rename("omega_d#{s.name}_dp")
    w_ds_dp.long_name = "Vertical Advection of #{s.name}"

    q1 = ds_dt + div_s + w_ds_dp

    return q1, ds_dt, div_s, w_ds_dp
  end

  #-- calc dry static energy
  # input
  # t  : 4D GPhys class, temperature
  # gph: 4D GPhys class, geopotential height
  # sh : 4D GPhys class, specific humidity
  # return
  # s  : 4D GPhys class, dry static stability
  def calc_dry_static_energy( t, gph, sh )
    cp = Cp * (1.0+0.94*sh)    # moist isobaric specific heat
    s = cp * t + gph
    s.rename("s")
    s.long_name = "dry static energy"

    return s
  end

  #-- calc Q1 (apparent heat source)
  def calc_q1( s, u, v, omega )

    q1, ds_dt, div_s, w_ds_dp = total_differentiation( s, u, v, omega )
    q1.rename("q1")
    q1.long_name = "apparent heat source"

    return q1, ds_dt, div_s, w_ds_dp
  end

  def derive_q1( t, gph, sh, u, v, omega )
    t = prep_deviation( t, "t" )
    u = prep_deviation( u, "u")
    v = prep_deviation( v, "v" )
    gph = prep_deviation( gph, "gph" )
    omega = prep_deviation( omega, "omega" )
    sh = prep_deviation( sh, "specific_humidity" )

    s = calc_dry_static_energy( t, gph, sh )
    q1, ds_dt, div_s, w_ds_dp = calc_q1( s, u, v, omega )

    return q1, s, ds_dt, div_s, w_ds_dp
  end

  #-- calc Q2 (apparent moisture sink)
  def calc_q2( sh, u, v, omega )

    q2_dum, ds_dt, div_s, w_ds_dp = total_differentiation( sh, u, v, omega )
    q2 = -Lc*q2_dum
    q2.rename("q2")
    q2.long_name = "apparent moisture sink"

    return q2, ds_dt, div_s, w_ds_dp
  end

  def derive_q2( sh, u, v, omega )

    sh = prep_deviation( sh, "specific_humidity" )
    u = prep_deviation( u, "u")
    v = prep_deviation( v, "v" )
    omega = prep_deviation( omega, "omega" )

    q2, ds_dt, div_s, w_ds_dp = calc_q2( sh, u, v, omega )

    return q2, ds_dt, div_s, w_ds_dp
  end

end

#================
if $0 == __FILE__

  year = ARGV[0].to_i
  year = 1979 if year == 0

  dir = "~/data/NetCDF/era_interim/4times/2.5x2.5/"
  fname_tail = "_#{year}.nc"
  t = GPhys::IO.open(dir+"t#{fname_tail}","t")
  u = GPhys::IO.open(dir+"u#{fname_tail}","u")
  v = GPhys::IO.open(dir+"v#{fname_tail}","v")
  gph = GPhys::IO.open(dir+"gph#{fname_tail}","gph")
  omega = GPhys::IO.open(dir+"omega#{fname_tail}","omega")
  sh = GPhys::IO.open(dir+"sh#{fname_tail}","sh")
  p dir+"t#{fname_tail}"

  #-- outfile
  outdir = "./"
  outfname = "q1q2_#{year}.nc"
  p outfname

  #==== calc Q1 (apparent heat source)
print "calculating Q1 (apparent heat source) ...\n"
  q1, s, ds_dt, div_s, w_ds_dp \
    = GPhys_Function::derive_q1( t, gph, sh, u, v, omega )
  #==== calc Q2 (apparent moisture sink)
print "calculating Q2 (apparent moisture sink) ...\n"
  q2, dsh_dt, div_sh, w_dsh_dp \
    = GPhys_Function::derive_q2( sh, u, v, omega )

  #=== output NetCDF
  outfile = NetCDF.create(outdir+outfname)
  GPhys::NetCDF_IO.write( outfile, s )
  GPhys::NetCDF_IO.write( outfile, q1 )
  GPhys::NetCDF_IO.write( outfile, ds_dt )
  GPhys::NetCDF_IO.write( outfile, div_s )
  GPhys::NetCDF_IO.write( outfile, w_ds_dp )
  GPhys::NetCDF_IO.write( outfile, q2 )
  GPhys::NetCDF_IO.write( outfile, dsh_dt )
  GPhys::NetCDF_IO.write( outfile, div_sh )
  GPhys::NetCDF_IO.write( outfile, w_dsh_dp )
  outfile.close
end

